/* > sht.c */
/* (C) 2005, Daniel F. Smith <dfs1122@gmail.com> */
/*
 * This is free software covered by the Lesser GPL.
 * See COPYING or http://gnu.org/licenses/lgpl.txt for details.
 * If you link your binaries with this library you are obliged to provide
 * the source to this library and your modifications to this library.
 */

/* Controls the Sensirion SHT1x/SHT7x temperature / relative humidity probe */
/* See http://sensirion.com for details */
/* For Linux, uses direct bit-bashing to a bidirectional parallel port */
/* Needs root privilage for I/O access */
/* You need the I/O address of the bidirectional parallel port */

/* Connect 5V   to power                 */
/* Connect 0V   to gnd     (DB25 pin 18--25) */
/* Connect clk  to -strobe (DB25 pin 1) or -autolf (DB25 pin 14) */
/* Connect data to dX      (DB25 pin 2--9) */
/* The data line needs a pull up resistor to 5V (10kOhm) */
/* To use the sensor's heater, you will need external 5V @ >8mA */

/* 
 * Run with:         gcc -Wall -O2 -Dtestsht sht.c -o test && ./test
 * Make object with: gcc -Wall -O2 -c sht.c
 */

#include <stdio.h>  /* for printf/fprintf/vfprintf/stderr */
#include <stdarg.h> /* for va_* used by trace function */
#include <stdlib.h> /* for calloc/free/NULL */
#include <unistd.h> /* for sleep/usleep */
#include <sys/io.h> /* for inb/outb used by pp_* functions */
#include "sht.h"

/* -- debug constructs for tracing and visual tracing -- */

#if 0
static int trace(const char *fmt,...) {
	int r;
	va_list va;
	va_start(va,fmt);
	r=vfprintf(stderr,fmt,va);
	fflush(stderr);
	va_end(va);
	return r;
	}

static int scope(int d,int c) {
	#define SCOPEMAX 1024
	static char dd[SCOPEMAX],cc[SCOPEMAX];
	static int lastd,lastc;
	static int t=0;
	const char steady[2]="_~",change[2]="\\/";

	if (d>=0 && c>=0 && t<SCOPEMAX-3) {
		d=!!d;
		c=!!c;
		dd[t]=(lastd==d)?steady[d]:change[d];
		cc[t]=(lastc==c)?steady[c]:change[c];
		t++;
		dd[t]=steady[d];
		cc[t]=steady[c];
		t++;
		lastd=d;
		lastc=c;
		if (t<SCOPEMAX-1) return t;
		}
	/* print trace */
	dd[t]='\0';
	cc[t]='\0';
	trace("D: %s\n",dd);
	trace("C: %s\n\n",cc);
	lastd=!!d;
	lastc=!!c;
	return t=0;
	}
#else
static int trace(const char *fmt,...) {return 0;}
static int scope(int d,int c) {return 0;}
#endif
#define TRACE trace
#define SCOPE scope

/* -- parallel port functions -- */

/* #include <sys/io.h> */
#define PP_REGISTERS 3
typedef struct parallelport_s parallelport;
struct parallelport_s {
	parallelport *next; /* next in list */
	unsigned int addr;  /* I/O address */
	unsigned int shadow[PP_REGISTERS]; /* local copy of registers */
	int count;          /* number of users of this structure */
	};
static parallelport *pp_baseport=NULL;

/* control bits */
#define PP_NSTB    0x01
#define PP_NAUTOLF 0x02
#define PP_NINIT   0x04
#define PP_NSELECT 0x08
#define PP_IRQEN   0x10
#define PP_NOUTIN  0xE0
/* status bits */
#define PP_NERROR  0x08
#define PP_SELECT  0x10
#define PP_NPAPER  0x20
#define PP_NACK    0x40
#define PP_NBUSY   0x80
/* program registers */
#define pp_ioout(x,a) outb(x,a)
#define pp_ioin(a) inb(a)
#define pp_andxor(shadow,and,xor) ((shadow)=(((shadow) & (and)) ^ (xor)))
#define pp_out(p,r,a,x) pp_ioout(pp_andxor((p)->shadow[r],a,x),(p)->addr+(r))
#define pp_in(p,r)      ((p)->shadow[r]=pp_ioin((r)+(p)->addr))
#define pp_contout(p,a,x) pp_out(p,2,a,x)
#define pp_contin(p)      pp_in(p,2)
#define pp_contshadow(p)  ((p)->shadow[2]) /* is write only register */
#define pp_statout(p,a,x) pp_out(p,1,a,x)
#define pp_statin(p)      pp_in(p,1)
#define pp_dataout(p,a,x) pp_out(p,0,a,x)
#define pp_datain(p)      pp_in(p,0)
#define pp_datashadow(p)  ((p)->shadow[0]) /* can be used as lvalue */
/* handy defines */
#define pp_datadir(p,w)  pp_contout(p,~PP_NOUTIN, (w)?0:PP_NOUTIN)
#define pp_getdatadir(p) (!(pp_contshadow(p)&PP_NOUTIN))
#define pp_strobe(p,x)   pp_contout(p,~PP_NSTB,   (x)?0:PP_NSTB)
#define pp_autolf(p,x)   pp_contout(p,~PP_NAUTOLF,(x)?0:PP_NAUTOLF)
#define pp_init(p,x)     pp_contout(p,~PP_NINIT,  (x)?0:PP_NINIT)
#define pp_select(p,x)   pp_contout(p,~PP_NSELECT,(x)?0:PP_NSELECT)
#define pp_contnout(p,a,x) pp_contout(p,a,(x)^(PP_NINIT|PP_NSELECT|PP_NAUTOLF|PP_NSTB))
/* allow I/O register access */
#define pp_ioon(p) ( ioperm((p)->addr,PP_REGISTERS,1)?-1:\
			(pp_contin(p),pp_statin(p),pp_datain(p),0) )
#define pp_iooff(p)  ioperm((p)->addr,PP_REGISTERS,0)

static parallelport *pp_getnextport(parallelport *p) {
	if (!p) return pp_baseport;
	return p->next;
	}

static void pp_close(parallelport *p) {
	if (!p) return;
	p->count--;
	if (p->count>0) return;

	/* remove port from list and delete it */
	(void)pp_iooff(p);
	if (pp_baseport==p)
		pp_baseport=p->next;
	else {
		parallelport *prev;
		for(prev=pp_baseport;prev && prev->next!=p;prev=prev->next);
		if (prev && prev->next==p) prev->next=p->next;
		}
	TRACE("Removed parallel port at 0x%04X\n",p->addr);
	free(p);
	}

static parallelport *pp_open(unsigned int ioaddr) {
	parallelport *p;
	
	/* find existing port */
	for(p=pp_baseport;p;p=p->next)
		if (p->addr==ioaddr) break;
		
	if (!p) {
		/* need new parallel port object */
		p=calloc(1,sizeof(*p));
		if (!p) return NULL;
		p->addr=ioaddr;
		if (pp_ioon(p)<0) {free(p); return NULL;}

		/* add to port list */
		p->next=pp_baseport;
		pp_baseport=p;
		TRACE("Added parallel port at 0x%04X\n",p->addr);
		}
	p->count++;
	return p;
	}

/* -- CRC8 and bit reverse -- */

typedef unsigned char u8;

static int crc8(int crc,int x) {
	static u8 *table=NULL;
	if (crc<0 || x<0) {
		TRACE("crc8 cleanup\n");
		if (table) free(table);
		table=NULL;
		return -1;
		}
	if (!table) {
		/* front-load calculation */
		const int gen=0x131;
		int i,j,r,s;
		table=calloc(256,sizeof(*table));
		TRACE("Creating crc8 table%s\n",table?"":" (fallback mode)");
		if (!table) {i=crc&0xFF; r=x;} else {i=0; r=0;}
		while(i<256) {
			s=i<<1;
			for(j=0;j<8;j++) {
				r<<=1; if ((s^r)&0x100) r^=gen;
				s<<=1;
				}
			if (!table) return r;
			table[i++]=r;
			r=0;
			}
		if (!table) return -1; /* can't get here */
		}
	return table[(crc ^ x)&0xFF];
	}

static int rev8(int x) {
	static u8 *table=NULL;
	if (x<0) {
		TRACE("rev8 cleanup\n");
		if (table) free(table);
		table=NULL;
		return -1;
		}
	if (!table) {
		/* front-load calculation */
		int i,j,r,s;
		table=calloc(256,sizeof(*table));
		TRACE("Creating rev8 table%s\n",table?"":" (fallback mode)");
		for(i=table?0:x;i<256;i++) {
			r=0; s=i;
			for(j=0;j<8;j++) {
				r<<=1; r|=s&1;
				s>>=1;
				}
			if (!table) return r;
			table[i]=r;
			}
		}
	return table[x&0xFF];
	}

/* -- SHT functions --- */

struct shtport_s {
	/* mostly private to I/O */
	parallelport *port;   /* which port we're connected to */
	unsigned int dmask;   /* which data line we're connected to */
	unsigned int cmask;   /* which clock line we're connected to */
	/* open */
	unsigned int tmpcmask;/* group cmask when linked through next */
	shtport *next;        /* used when we have a list of SHT devices */
	unsigned long int lastdata; /* the last data bits read from this SHT (not CRC) */
	unsigned int lastack; /* the last ack bits (0=good, shifted in from R) */
	unsigned int crc;     /* the last CRC calculation/syndrome (0=good) */
	unsigned int tmp;     /* temporary storage */
	double tempoffset;    /* linear value to add to temperature reading */
	double lasttemp;      /* last temp value read */
	double lastrh;        /* last RH value read */
	}; /* typedef is in header */

/* sht commands */
#define sht_cgettemp   0x03
#define sht_cgetrh     0x05
#define sht_cgetstatus 0x07
#define sht_csetstatus 0x06
#define sht_csetreset  0x1E

#define sht_statusval 0x00 /* what we expect the status register/flags to be */
#define sht_statuslowbatt 0x40 /* status flag reporting low battery */
#define sht_clocktimeus 50 /* half-clock cycle time in microseconds */
#define sht_readingclocksmax (40000/(2*sht_clocktimeus)) /* timeout for wait */

/* interface level control */
	/* read clk line, set clk high, set clk low */
#define clkr(s)  (!( pp_contshadow((s)->port) & (s)->cmask ))
#define clkh(s)  pp_contnout((s)->port,~(s)->tmpcmask,(s)->tmpcmask)
#define clkl(s)  pp_contnout((s)->port,~(s)->tmpcmask,0)
	/* read data line, set data line high or low */
#define dr(s)    (!!(pp_datashadow((s)->port)&(s)->dmask))
#define dw(s,x)  pp_andxor(pp_datashadow((s)->port),~((s)->dmask),(x)?(s)->dmask:0)
	/* set data direction to inb (read) or outb (write) */
#define ddin(s)  pp_datadir((s)->port,0)
#define ddout(s) pp_datadir((s)->port,1)
	/* wait a half-clock with data I/O */
#define w(s) do {\
	if (pp_getdatadir((s)->port)==1) pp_dataout((s)->port,0xFF,0x00);\
	usleep(sht_clocktimeus);\
	if (pp_getdatadir((s)->port)==0) pp_datain((s)->port);\
	SCOPE(!!dr(s),clkr(s));\
	} while(0)
	/* generalize misc and port open/close */
#define power(s,on) /* energize 5V line */
#define openport(ioaddr) pp_open(ioaddr)
#define closeport(p) pp_close(p)
#define remainingports() (!!pp_getnextport(NULL))
	/* nothing below here should be using the pp_* functions directly */

/* functions that might be exported sometime */

/* conversions and constants */

static double sht_temp14offset(double voltage) {
	if      (voltage>=4.0) return -39.75 - 0.25*(voltage-4.0);
	else if (voltage>=3.5) return -39.66 - 0.18*(voltage-3.5);
	else if (voltage>=3.0) return -39.60 - 0.12*(voltage-3.0);
	else                   return -39.55 - 0.10*(voltage-2.5);
	}

static double sht_temp14(shtport *s) {
	/* convert 14-bit temperature value to celsius with (milli) voltage correction */
	const double d2=0.01;
	if (!s) return d2;
	if (s->lastack || s->crc) return sht_badtemp;
	return (s->lastdata & 0xFFFF)*d2 + s->tempoffset;
	}

static double sht_rh12(shtport *s,double temp) {
	/* convert 12-bit rh value to % rh with temperature correction */
	const double c1=-4.0,c2=0.0405,c3=-2.8e-6;
	const double t1=0.01,t2=0.00008;
	double srh,rh;
	if (!s) return c2;
	if (s->lastack || s->crc) return sht_badrh;
	srh=(s->lastdata & 0xFFFF);
	rh=c1 + c2*srh + c3*srh*srh;
	if (temp>=-40.0 && temp!=sht_badtemp) /* temp correction */
		rh+=(temp-25.0)*(t1 + t2*srh);
	return rh;
	}

/* low level I/O */

static void sht_out8(shtport *s,unsigned int x) {
	/* send 8 bits of x to every device on SHT list, lastack&1 will be 0 if acked */
	shtport *ss;
	TRACE("out8 %02X\n",x);
	ddout(s);
	clkl(s);
	/* data is good on rising edge of clk */
	for(x=0x00100 | (x&0x000FF);!(x&0x10000);x<<=1) {
		for(ss=s; ss; ss=ss->next) dw(ss,x&128);
		w(s); clkh(s); w(s); clkl(s);
		}
	/* ack */
	ddin(s);
	w(s);
	for(ss=s; ss; ss=ss->next) {
		ss->crc=crc8(ss->crc,(x>>8)&0xFF);
		ss->lastack=(ss->lastack<<1) | dr(ss);
		}
	clkh(s); w(s); clkl(s);
	}

static void sht_in8(shtport *s,int last,int crc) {
	/* read 8 bits of data from SHT list, set last to prevent ack, crc indicates CRC input */
	int i;
	shtport *ss;

	ddin(s);
	clkl(s);
	w(s);
	for(i=0;i<8;i++) {
		for(ss=s; ss; ss=ss->next) ss->tmp=(ss->tmp<<1) | dr(ss);
		clkh(s); w(s); clkl(s); w(s);
		}

	/* ack and CRC calculation */
	for(ss=s; ss; ss=ss->next) {
		dw(ss,last);
		ss->tmp&=0xFF;
		TRACE("in8 mask=0x%X data=0x%02X\n",ss->dmask,ss->tmp);
		if (crc)
			ss->crc=crc8(ss->crc,rev8(ss->tmp));
		else {
			ss->crc=crc8(ss->crc,ss->tmp);
			ss->lastdata=(ss->lastdata<<8)|ss->tmp;
			}
		}
	ddout(s);
	w(s); clkh(s); w(s); clkl(s);
	ddin(s);
	w(s);
	}

static void sht_wait(shtport *s) {
	/* wait for all SHTs to finish reading */
	/* the lastack will be 0 if waiting succeeded */
	int t=0;
	int tracking=0;
	shtport *ss;

	TRACE("waiting...");
	ddin(s);
	clkl(s);
	for(ss=s; ss; ss=ss->next) {
		ss->tmp=!(ss->lastack&1); /* 0 indicates finished/error */
		tracking+=ss->tmp;
		}
	for(t=0;t<sht_readingclocksmax;t++) {
		w(s);
		for(ss=s; ss; ss=ss->next) {
			if (!ss->tmp) continue;
			if (dr(ss)!=0) continue;
			ss->tmp=0;
			tracking--;
			}
		if (tracking<=0) break;
		}
	for(ss=s; ss; ss=ss->next)
		ss->lastack=(ss->lastack<<1) | ss->tmp;
	TRACE(" %d clocks%s\n",t,t>=sht_readingclocksmax?"(timeout) ":"");
	}

static int sht_start(shtport *s,int fast) {
	/* send reset communications and then start sequence to sht */
	const u8 startseq[]={ /* clk is b0, data is b1 */
		4,
		2,3,2,3,2,3,2,3,2,3,2,3,2,3,2,3,2,3, /* comm. reset */
		16, /* stop fast skip */
		4, /* ddout */
		2,3,1,0,1,3,2, /* start */
		8,128 /* finish sequence */};
	const u8 *q;
	shtport *ss;

	/* initialize SHT list */
	for(ss=s; ss; ss=ss->next) {
		if (ss->lastack) fast=0; /* errors prevalent---don't do fast start */
		ss->crc=rev8(sht_statusval&0x0F);
		ss->lastack=0;
		}

	TRACE("start\n");
	ddin(s);
	for(q=startseq;;q++) {
		if ((*q)&128) break;
		if ((*q)&16) {fast=0; continue;}
		if (fast) {continue;}
		if ((*q)&8) {ddin(s); continue;}
		if ((*q)&4) {ddout(s); continue;}

		/* set clk from b0 and data from b1 */
		for(ss=s; ss; ss=ss->next) dw(ss,(*q)&2);
		if ((*q)&1) clkh(s); else clkl(s);
		w(s);
		}
	return 0;
	}

/* mid-level commands */

static void sht_cmdwrite(shtport *s,int cmd,...) {
	/* send commands and bytes to SHT list (with comm. reset) */
	va_list va;
	TRACE("write cmd=%u\n",cmd);
	sht_start(s,0);  SCOPE(-1,-1);
	sht_out8(s,cmd); SCOPE(-1,-1);
	sht_wait(s);     SCOPE(-1,-1);
	va_start(va,cmd);
	while((cmd=va_arg(va,int))>=0) {
		sht_out8(s,cmd); SCOPE(-1,-1);
		}
	va_end(va);
	/* lastack&3 should be 0 */
	}	

static void sht_cmdread(shtport *s,int cmd,int n) {
	/* send a command and input a byte from SHT list */
	TRACE("read1 cmd=0x%02X n=%d\n",cmd,n);
	sht_start(s,1);  SCOPE(-1,-1);
	sht_out8(s,cmd); SCOPE(-1,-1);
	sht_wait(s);     SCOPE(-1,-1);
	while(n-->0) {
		sht_in8(s,0,0); SCOPE(-1,-1);
		}
	sht_in8(s,1,1);   SCOPE(-1,-1); /* CRC */
	/* lastack&3 should be 0, crc should be 0 */
	}

/* -- exported functions -- */

void sht_close(shtport *s) {
	/* close down the sht data structures */
	if (s) {
		if (s->port) closeport(s->port);
		free(s);
		}

	/* clean up internal tables if nothing left */
	if (!remainingports()) {
		crc8(-1,-1);
		rev8(-1);
		}
	}

shtport *sht_open(unsigned int ioaddr,int dmask,int cmask,double voltage,int *errcode) {
	/* get a handle to sht device at given I/O address and bit location */
	/* parallel port must be statically available (global) */
	shtport *s=NULL;
	int errc=-1; /* error code */
	
	TRACE("open\n");
	do {
		/* make new sht object */
		s=calloc(1,sizeof(*s));
		if (!s) {errc=1; break;}
		s->next=NULL;
		s->tempoffset=sht_temp14offset(voltage);
		s->lasttemp=sht_badtemp;
		s->lastrh=sht_badrh;
		s->lastack=1; /* force communication reset */
		s->dmask=dmask;
		s->cmask=s->tmpcmask=cmask;
		s->port=openport(ioaddr);
		if (!s->port) {TRACE("could not open port\n"); errc=2; break;}

		/* at this point the shtport is usable */
		clkl(s); /* clk should be low at power on */
		dw(s,0);
		power(s,1); /* ensure power on: raise 5V line */
		w(s);
		TRACE("sht at dmask=0x%02X cmask=0x%02X",s->dmask,s->cmask);

		for(;;) {
			/* check status */
			sht_cmdread(s,sht_cgetstatus,1);
			TRACE("sht_open sht_getstatus ack=0x%X status=0x%02X crc=0x%02X\n",
				s->lastack,(s->lastdata&0xFF),s->crc);
			if (s->lastack) {errc=3; break;}
			if ((s->lastdata&0x0F)==(sht_statusval&0x0F)) {
				/* CRC only correct with expected status value */
				if (s->crc) {errc=4; break;}
				}
			if (s->lastdata & sht_statuslowbatt) {
				/* check low battery */
				if (errc>0) break;
				TRACE("low battery\n");
				/* try power off, on then retry */
				power(s,0);
				sleep(2);
				power(s,1);
				sleep(1);
				errc=5;
				continue;
				}
			if ((s->lastdata&0xFF)!=sht_statusval) {
				/* status is unusual---reset device, set status and try again */
				if (errc>0) break;
				TRACE("status strange: resetting\n");
				sht_cmdwrite(s,sht_csetreset,-1);
				if (s->lastack) {errc=3; break;}
				sht_cmdwrite(s,sht_csetstatus,sht_statusval,-1);
				if (s->lastack) {errc=3; break;}
				errc=6;
				continue;
				}
			/* must be okay (recovered) if we got here */
			errc=0;
			break;
			}
		if (errc!=0) break;
		/* good exit */
		SCOPE(-1,-1);
		return s;
		} while(0);
	/* cannot find SHT: error */
	if (errcode) *errcode=errc;
	SCOPE(-1,-1);
	sht_close(s);
	return NULL;
	}

int sht_readmany(shtport *s,...) {
	/* read temp and RH from many SHT devices on the same port */
	int errs=0;
	shtport *ss;
	va_list va;
	
	if (!s) return -1;
	/* link shtport list */
	ss=s;
	s->tmpcmask=s->cmask;
	va_start(va,s);
	do {
		ss->next=va_arg(va,shtport*);
		s->tmpcmask|=ss->cmask;
		ss=ss->next;
		if (!ss) break;
		if (ss->port != s->port) {errs=-1; break;}
		} while(ss);
	va_end(va);
	if (errs) return errs;

	/* read temps */
	TRACE("readmany: temp\n");
	sht_cmdread(s,sht_cgettemp,2);
	for(ss=s; ss; ss=ss->next) {
		ss->lasttemp=sht_temp14(ss);
		if (ss->lasttemp==sht_badtemp) errs++;
		}

	/* read RHs */
	TRACE("readmany: rh\n");
	sht_cmdread(s,sht_cgetrh,2);
	for(ss=s; ss; ss=ss->next) {
		ss->lastrh=sht_rh12(ss,ss->lasttemp);
		if (ss->lastrh==sht_badrh) errs++;
		}
	return errs;
	}
	
void sht_getgranularity(shtport *s,double *temp,double *rh) {
	if (temp) *temp=sht_temp14(NULL);
	if (rh) *rh=sht_rh12(NULL,sht_badtemp);
	}
void sht_getlasttemprh(shtport *s,double *temp,double *rh) {
	if (!s) {if (temp) *temp=sht_badtemp; if (rh) *rh=sht_badrh;}
	if (temp) *temp=s->lasttemp;
	if (rh) *rh=s->lastrh;
	}
double sht_getlasttemp(shtport *s) {return s?s->lasttemp:sht_badtemp;}
double sht_gettemp(shtport *s) {
	if (!s) return sht_badtemp;
	s->next=NULL;
	s->tmpcmask=s->cmask;
	sht_cmdread(s,sht_cgettemp,2);
	return s->lasttemp=sht_temp14(s);
	}
double sht_getlastrh(shtport *s) {return s?s->lastrh:sht_badtemp;}
double sht_getrh(shtport *s,double temp) {
	if (!s) return sht_badrh;
	s->next=NULL;
	s->tmpcmask=s->cmask;
	sht_cmdread(s,sht_cgetrh,2);
	return s->lastrh=sht_rh12(s,temp);
	}

/* -- for testing -- */
#ifdef testsht

#include <time.h> /* for time/difftime/time_t */

int main(void) {
	double temp,rh;
	shtport *s=NULL;
	const char *err=NULL;
	int errc=-1;
	time_t starttime;
	#define elapsed() difftime(time(NULL),starttime)
	time(&starttime);

	#if 0 /* check voltage/temp calculation: local test only */
	printf("# Voltage\ttemp offset\n");
	for(temp=2.0;temp<5.5;temp+=0.1)
		printf("%.2f\t%.3f\n",temp,sht_temp14offset(temp));
	#endif
	
	do {
		printf("Opening SHT\n");
		s=sht_open(PARALLEL_PORT_IO_ADDR_0,1<<1,PARALLEL_PORT_NSTB,5.0,&errc);
		if (!s) {err="no SHT found"; break;}
		
		/* print temp */
		temp=sht_gettemp(s);
		if (temp==sht_badtemp) err="bad temp reading";
		printf("Temp=%.2fC (%.1fF)\n",temp,1.8*temp+32.0);
		
		/* print RH */
		rh=sht_getrh(s,temp);
		if (rh<0.0) err="bad RH reading";
		printf("RH=%.1f%%\n",rh);

		/* print time taken */
		printf("time_elapsed=%.1fs\n",elapsed());
		
		if (err) break;
		
		#if 1
		/* print a table */
		printf("\n# time/s\ttemp/C\trh/%%\n");
		while(elapsed()<100) {
			sleep(10-(int)elapsed()%10); /* print on ~10s marks */
			if (sht_readmany(s,NULL)!=0) {err="sht_read failed"; break;}
			printf("%.0f\t%.2f\t%.1f\n",
				elapsed(),sht_getlasttemp(s),sht_getlastrh(s));
			}
		if (err) break;
		#endif
		} while(0);
	if (s) sht_close(s);

	if (err) fprintf(stderr,"Error reported: %s (code %d)\n",err,errc);
	return err?1:0;
	}
#endif /* defined testsht */
